use anyhow::{Result,Context};
use std::{fs}; //,io};
use std::fs::OpenOptions;
use std::io::{Write};
use std::path::Path;
use std::str;
// use std::ffi::OsString;

use reqenv::ReqEnv;
use crate::clouds::defs::{
	Cloud,
	Provider,
	MainResourcesConfig,
	HostInfo,
};
use crate::clouds::defs::{SrvcsHostInfo};
use crate::clouds::upcloud::{run_on_upcloud_info,run_on_upcloud_check};
use crate::liveness::{ create_cloud_check };
use crate::clouds::utils::{
	load_cloud_env,
	load_cloud_name_config,
	get_cloud_home_list,
	create_cloud_config,
	cloud_info_json,
	write_cache_file,
};

pub async fn run_on_provider_info(
		reqname: &str, reqenv: &ReqEnv, req_tsksrvc: &str, req_srvrs: &str,
		provider: Provider, source: &str, cfg_data: String,
		cld_name: &str,
		env_cloud: &Cloud
) -> Vec<HostInfo> {
	match provider.name.as_str() {
		"upcloud" => {
			// TODO clean SSH keys or encrypt content
			// dbg!(&cloud_config);
			run_on_upcloud_info(reqname,&reqenv,req_tsksrvc,req_srvrs, source, cfg_data, &cld_name,env_cloud).await
		},
		_ => {
			let result = format!("Errors on {} provider {} not found",&source,&provider.name);
			if envmnt::get_isize("DEBUG",0) > 1 {
				println!("{}",&result);
			}
			Vec::new()
		}
	}
}
pub async fn run_on_provider_check(
		reqname: &str, req_tsksrvc: &str, req_srvrs: &str,
		provider: Provider, source: &str, cfg_data: String,
		env_cloud: &Cloud
) -> Vec<SrvcsHostInfo> {
	match provider.name.as_str() {
		"upcloud" => {
			// TODO clean SSH keys or encrypt content
			// dbg!(&cloud_config);
			run_on_upcloud_check(reqname,req_tsksrvc,req_srvrs, source, cfg_data, env_cloud).await
		},
		_ => {
			if envmnt::get_isize("DEBUG",0) > 1 {
			  println!("Errors on {} provider {} not found",&source,&provider.name);
			}
			Vec::new()
		}
	}
}
pub async fn on_cloud_name_req_info(
	reqname: &str,env_cloud: &Cloud,reqenv: &ReqEnv,
	req_tsksrvc: &str, req_srvrs: &str, cld_name: &str, source: &str
) -> Vec<HostInfo> {
	let mut cloud = env_cloud.to_owned();
	load_cloud_env(&mut cloud, &source).await
		.unwrap_or_else(|e| { 
			eprintln!("load_cloud_env: {}",e);
	});
	let (cfg,provider,cfg_data) = load_cloud_name_config(&mut cloud, &source).await
		.unwrap_or_else(|e| {
			eprintln!("load_cloud_name_config: {}",e);
			(MainResourcesConfig::default(),Provider::default(),String::from(""))
	});
	if cfg.mainName.is_empty() || cfg_data.is_empty() {
		let result = format!("Errors loading {}",&source);
		if envmnt::get_isize("DEBUG",0) > 1 {
			println!("{}",&result);
		}
		return Vec::new();
	}
	run_on_provider_info(&reqname,&reqenv,&req_tsksrvc,&req_srvrs,provider,&source,cfg_data,&cld_name,&cloud).await
}
pub async fn on_cloud_name_req_check(
	reqname: &str,env_cloud: &Cloud,_reqenv: &ReqEnv,
	req_tsksrvc: &str, req_srvrs: &str, source: &str
) -> Vec<SrvcsHostInfo> {
	let mut cloud = env_cloud.to_owned();
	load_cloud_env(&mut cloud, &source).await
		.unwrap_or_else(|e| { 
			eprintln!("load_cloud_env: {}",e);
	});
	let (cfg,provider,cfg_data) = load_cloud_name_config(&mut cloud, &source).await
		.unwrap_or_else(|e| {
			eprintln!("load_cloud_name_config: {}",e);
			(MainResourcesConfig::default(),Provider::default(),String::from(""))
	});
	if cfg.mainName.is_empty() || cfg_data.is_empty() {
		if envmnt::get_isize("DEBUG",0) > 1 {
		  println!("Errors loading {}",&source);
		}
		return Vec::new();
	}
	run_on_provider_check(&reqname,&req_tsksrvc,&req_srvrs,provider,&source,cfg_data,&cloud).await
}

pub async fn on_cloud_req(
	reqname: &str,env_cloud: &Cloud,reqenv: &ReqEnv,
	req_tsksrvcs: &str,_req_srvrs: &str, source: &str
) -> String {
	let config = reqenv.config();
	let debug = envmnt::get_isize("DEBUG",0);
  // let lock_path = format!("{}/{}_{}{}",&config.cache_lock_path,&reqname,&req_tsksrvcs.replace(",","_"),&config.cache_lock_ext);
	// if Path::new(&lock_path).exists() || reqname.ends_with("_job") {
	let output_path = format!("{}/{}_{}.json",&config.cache_path,&reqname,&req_tsksrvcs.replace(",","_"));
	if ! reqname.ends_with("_job") {
		println!("{}",&output_path);
		if Path::new(&output_path).exists() {
			if debug > 0 {
				println!("Using cache: {} at {}",&output_path,envmnt::get_or(format!("LAST_CACHE_{}",&output_path), ""));
			}
			return fs::read_to_string(&output_path).with_context(|| format!("Failed to read json cache 'outut_path' from {}", &output_path))
				.unwrap_or_else(|e| { 
					eprintln!("read file {}: {}",&output_path,e);
					String::from("")
			}).to_owned();
		} 
	}
	let lock_path = format!("{}/{}_{}.{}",&config.cache_lock_path,&reqname,&config.cache_items.replace(",","_"),&config.cache_lock_ext);
	if Path::new(&lock_path).exists() { 
		if envmnt::get_or(format!("LAST_CACHE_{}",&output_path),"") != "" {
			if debug > 0 {
				println!("{}: lock found {} ",&reqname,&lock_path);
			}
			// return Err(anyhow!("Lock found {} ",&lock_path));		
			return fs::read_to_string(&output_path).with_context(|| format!("Failed to read json cache 'outut_path' from {}", &output_path))
				.unwrap_or_else(|e| { 
					eprintln!("read file {}: {}",&output_path,e);
					String::from("")
		  }).to_owned();
		  // } else {
		  // 	println!("Not LAST_CACHE environment found for lock: {} ",&lock_path);
		}
  }
	let mut cloud = env_cloud.to_owned();
	load_cloud_env(&mut cloud, &source).await
		.unwrap_or_else(|e| { 
			eprintln!("load_cloud_env {}",e);
	});
	let entries: Vec<String>;
	if source == "*" {
		entries = get_cloud_home_list(&cloud).await
		.unwrap_or_else(|e| { 
			eprintln!("get_cloud_home_list: {}",e);
			Vec::new()
		});
	} else {
		entries = vec!(source.to_string());
	}
	match reqname  {
	"liveness_job" | "apps_job" | "status_job" => { 
      serde_json::to_string(
				&create_cloud_check(req_tsksrvcs,reqenv,entries,cloud).await
			).unwrap_or_else(|_| String::from("")).replace("\n","")
		}
	_ => {  
	  let result = cloud_info_json(create_cloud_config(reqname,req_tsksrvcs,reqenv,entries,cloud.to_owned()).await);
		if !result.is_empty() {
			let _ = write_cache_file(&output_path.replace("_job",""),&result).unwrap_or_else(|e|{println!("Error writing {}:{}",&output_path,e)});
		}
		result
		}
	}
}
// ) -> BTreeMap<String,String> {
// 	let mut output_data =  BTreeMap::new();
// 	let mut req_tsksrvcs_items = req_tsksrvcs.split(',');
// 	for it in req_tsksrvcs_items.by_ref() {
// 	};
// }
pub async fn get_cloud_cache_req(reqenv: &ReqEnv,cloud: &Cloud, reqname: &str, reqname_job: &str) -> Result<()> {
	let debug = envmnt::get_isize("DEBUG",0); 
	if debug > 0 {
		println!("cloud cache {} ... {:?} ",reqname,chrono::Utc::now());
	}
	let config = reqenv.config();
	let lock_path = format!("{}/{}_{}.{}",&config.cache_lock_path,&reqname,&config.cache_items.replace(",","_"),&config.cache_lock_ext);
	let output_path = format!("{}/{}_{}.json",&config.cache_path,&reqname,&config.cache_items.replace(",","_"));
	if Path::new(&lock_path).exists() {
		if envmnt::get_or(format!("LAST_CACHE_{}",&output_path),"") != "" {
			if debug > 0 {
			  println!("Lock found {} ",&lock_path);
			}
     	// return Err(anyhow!("Lock found {} ",&lock_path));		
			return Ok(())
		} else {
		  println!("Not LAST_CACHE environment found for lock: {} ",&lock_path);
		}
	}
	let now = chrono::Utc::now().timestamp();
	envmnt::set(format!("LAST_CACHE_{}",&output_path), format!("{}",&now));
	let result = on_cloud_req(&reqname_job,&cloud,&reqenv,&config.cache_items,"","*").await;
  if Path::new(&output_path).exists() {
    fs::remove_file(&output_path)?;
  }
  let mut file = OpenOptions::new().write(true).create(true).open(&output_path)?;
  file.write_all(result.as_bytes())?;
  if debug > 0 {
 		println!("{}: [cloud config] -> {}\n",&now,&output_path);
	}
	Ok(())
}
pub async fn make_cloud_cache(reqenv: &ReqEnv,cloud: &Cloud) -> Result<()> {
	if envmnt::get_isize("DEBUG",0) > 0 {
	  println!("Making cloud cache {:?} ... ",chrono::Utc::now());
	}
  get_cloud_cache_req(reqenv, cloud, "config", "config_job").await.unwrap_or_else(|e| println!("Error cache config: {}",e));
	Ok(())
}
pub async fn run_cache_data(
	task: &str,reqenv: &ReqEnv,cloud: &Cloud,
	req_tsksrvcs: &str, source: &str
) -> Result<()> {
  // let out_data = get_cache_data(task,reqenv);
	// if out_data.is_empty() {
	let result = on_cloud_req(
		format!("{}_job",&task).as_str(),&cloud,&reqenv,
		task,req_tsksrvcs,source
	).await;
  write_cache_data(task,&result,reqenv,cloud).await
}
pub fn get_cache_info(task: &str,reqenv: &ReqEnv,debug: isize) -> String {
	if debug > 0 {
		println!("{} cache ... {:?} ",&task,chrono::Utc::now());
	}
	let config = reqenv.config();
	match task {
		"liveness" | "apps" | "status" => format!("{}/{}.json",&config.check_path,&task),
		_ => format!("{}/{}.json",&config.check_path,&task),
	}
}
pub async fn get_cache_data(task: &str,reqenv: &ReqEnv) -> String {
	let debug = envmnt::get_isize("DEBUG",0); 
	let output_path = get_cache_info(task,reqenv,debug);
	let output_data = fs::read_to_string(&output_path).with_context(|| format!("Failed to read json {} 'outut_path' from {}",&task,&output_path))
		.unwrap_or_else(|e| { 
			eprintln!("read file {}: {}",&output_path,e);
			String::from("")
	});
	if debug > 0 && !output_data.is_empty() {
    println!("{}: [{} cache] -> {}",chrono::Utc::now(),&task,&output_path);
	}
	output_data
}
pub async fn write_cache_data(task: &str,result: &str,reqenv: &ReqEnv,_cloud: &Cloud) -> Result<()> {
	let debug = envmnt::get_isize("DEBUG",0); 
	let output_path = get_cache_info(task,reqenv,debug);
	envmnt::set(format!("LAST_{}_CACHE{}",&task,&output_path), format!("{}",chrono::Utc::now()));
	//let result_map = on_cloud_req(format!("{}_job",&task).as_str(),&cloud,&reqenv,task,"","*").await;
	//if let Some(result) = result_map.get(task) {
	// let result = on_cloud_req(format!("{}_job",&task).as_str(),&cloud,&reqenv,task,"","*").await;
	if !result.is_empty() {
		// println!("{}",&output_path);
		if Path::new(&output_path).exists() {
			fs::remove_file(&output_path)?;
		}
		let mut file = OpenOptions::new().write(true).create(true).open(&output_path)?;
		file.write_all(result.as_bytes())?;
	} else {
    println!("{}: get data to write error",&task);
	}
	if debug > 0 {
    println!("{}: [{} cache] -> {}\n",chrono::Utc::now(),&task,&output_path);
	}
	Ok(())
}