use anyhow::{anyhow,Result,Context, Error};
use std::{fs}; //,io};
use std::path::Path;
use std::fs::OpenOptions;
use std::io::{Write};
use rfm::mkdir;
use std::str;
use std::process::{Command};
// use std::ffi::OsString;
use crate::clouds::defs::{
	CloudEnv,
	Cloud,
	Provider,
	SrvcsHostInfo,
	HostInfo,
	MainResourcesConfig,
};
use reqenv::ReqEnv;
use crate::defs::{
	KloudHome,
	KloudCheckHome,
	CloudGroup,
	CloudItem,
	SSHAccess,
};
use crate::clouds::upcloud::{get_upcloud_info};
use crate::monitor::get_cloud_monitor_info;
use crate::clouds::on_req::{
	on_cloud_name_req_check,
	on_cloud_name_req_info,
};

/// On_cloud
/// load __`item`__ form `envmnt` with `dflt`
/// Check if path from `source` exits for `item` or fallback to check from `root` path.
/// __`item`__ path is requiered to exist
pub async fn get_env_path(item: &str, dflt: &str, source: &str, root: &str, is_tpl: bool) -> Result<String> {
  let mut base = dflt.to_owned();
  if item.len() > 0 {
    base=envmnt::get_or(item,dflt);
  }
  if Path::new(&base).has_root() {
    if  Path::new(&base).exists() {
      return Ok(base);
    } else {
      return Err(anyhow!("Path {} not found", &base));
    }
  }
	#[allow(unused_assignments)]
  let mut item_path= String::from("");
  if source.len() > 0 {
    item_path = format!("{}/{}",&source,&base);
  } else {
    item_path = format!("{}",&base);
  }
  if ! Path::new(&item_path).exists() {
    item_path=format!("{}/{}",root,&base);
    if is_tpl && ! Path::new(&item_path).exists() {
      item_path=format!("{}/{}",&source,&base);
      if ! Path::new(&item_path).exists() {
        item_path=format!("{}/{}",&root,&base);
      }
    }
    if ! Path::new(&item_path).exists() {
      return Err(anyhow!("Path '{}' not found in {} - {}", &base,&source,&root));
    }
  }
  // println!("Path: {} src {} root {} item {}", &base, &source, &root, item_path);
	Ok(base)
}
/// env_cloud
/// Scanning environment from __CLOUD_PATH__ overloaded with __CLOUD_HOME__ `source` CliOpts argument (-s) in case
/// Load path from `envmnt`  and __`get_env_path`___
/// __`source`__ path is mandatory
pub async fn env_cloud(source: &str, cloud_env: &mut CloudEnv) -> Result<()> {
  cloud_env.path = envmnt::get_or("ROOT_KLDS", "");
	if cloud_env.path.is_empty() {
		return Err(anyhow!("Clouds Root Path {} not found", &cloud_env.source_path));		
	}
  cloud_env.home=get_env_path("KLDS_HOME","home", "", &cloud_env.path,false).await?;
  cloud_env.monitor_run=envmnt::get_or("KLD_MONITOR_RUN","kloud_mon");
  cloud_env.source=source.to_owned();
  cloud_env.config_root = envmnt::get_or("KLD_CONFIG_ROOT", "config");
	if source == "*" {
    cloud_env.config_path = envmnt::get_or("KLD_CONFIG", "config.yaml");
    cloud_env.config_json_path = envmnt::get_or("KLD_CONFIG_JSON", "config.json");
		return Ok(())
	}
  let arr_source: Vec<&str> = source.split_terminator("/").collect();
  if let Some(name) = arr_source.get(0) {
    cloud_env.cloud= format!("{}",&name);
  }
  if let Some(group) = arr_source.get(1) {
    cloud_env.group = format!("{}",&group);
  }
  if let Some(target) = arr_source.get(2) {
    cloud_env.target = format!("{}",&target);
  }
  cloud_env.source_path = format!("{}/{}",&cloud_env.home,&source);
  if ! Path::new(&cloud_env.source_path).exists() {
    return Err(anyhow!("Clouds Source Path {} not found", &cloud_env.source_path));
  }
  cloud_env.config_path=get_env_path("KLDS_CONFIG","config.yaml", 
  	format!("{}/{}",&cloud_env.source_path,&cloud_env.config_root).as_str(),
		&cloud_env.path,false).await?;
	cloud_env.config_json_path=get_env_path("KLDS_CONFIG_JSON","config.json", 
  	format!("{}/{}",&cloud_env.source_path,&cloud_env.config_root).as_str(),
		&cloud_env.path,false).await?;
  if cloud_env.group.is_empty() {
    cloud_env.provision=format!("{}/{}/{}/{}",&cloud_env.home,&cloud_env.cloud,envmnt::get_or("CLOUD_PROVISION","provision"),&cloud_env.target);
  } else {
    cloud_env.provision=format!("{}/{}/{}/{}/{}",&cloud_env.home,&cloud_env.cloud,&cloud_env.group,envmnt::get_or("CLOUD_PROVISION","provision"),&cloud_env.target);
  }
  if ! Path::new(&cloud_env.provision).exists() {
    let dir_path_buf = Path::new(&cloud_env.provision).to_path_buf();
    let dirs: Vec<&std::path::PathBuf> = vec![&dir_path_buf];
    mkdir(&dirs).with_context(|| format!("\nFailed to create dir path {}", &cloud_env.provision))?;
		if envmnt::get_isize("DEBUG",0) > 1 {
  	  println!("{} created", &cloud_env.provision);
		}
  }
  cloud_env.wk_path = envmnt::get_or("KLDS_WKDIR", "/tmp");
  cloud_env.tsksrvcs_path=get_env_path("KLDS_TSKSRVCS","tsksrvcs", &cloud_env.source_path,&cloud_env.path,false).await?;
  cloud_env.versions=get_env_path("KLDS_VERSIONS","versions.yaml", &cloud_env.source_path,&cloud_env.path,false).await?;
  Ok(())
}
pub async fn clear_specs(source: &str) -> Result<()> {
  let env_path = envmnt::get_or("ROOT_KLDS", "clouds");
  let env_home = get_env_path("KLDS_HOME","home", "", &env_path,false).await?;
  let env_source=format!("{}/{}",&env_home,&source);
  if ! Path::new(&env_source).exists() {
    return Ok(());
  }
  let env_provision=format!("{}/{}",&env_source,envmnt::get_or("KLDS_PROVISION","provision"));
  let env_specs=format!("{}/specs",env_provision);
  if Path::new(&env_specs).exists() {
		if envmnt::get_isize("DEBUG",0) > 1 {
	  	println!("Delete {}/specs",env_provision);
		}
    fs::remove_dir_all(&env_specs)?;
  }
  Ok(())
}
pub async fn load_config_data(cloud: &Cloud, source: &str) -> Result<String> {
	// dbg!(&cloud);
	let cfg_path = format!("{}/{}/{}/{}",&cloud.env.home,&source,&cloud.env.config_root,&cloud.env.config_path);
	let debug = envmnt::get_isize("DEBUG",0); 
	if debug > 1 {
		println!("load_config: {}",&cfg_path);
	}  
  let cfg_data = fs::read_to_string(&cfg_path).with_context(|| format!("Failed to read 'cfg_path' from {}", &cfg_path))?;
	Ok(cfg_data)
}
pub async fn load_config_json_data(cloud: &Cloud, source: &str) -> Result<String> {
	let cfg_path = format!("{}/{}/{}/{}",&cloud.env.home,&source,&cloud.env.config_root,&cloud.env.config_json_path);
  let cfg_data = fs::read_to_string(&cfg_path).with_context(|| format!("Failed to read json 'cfg_path' from {}", &cfg_path))?;
	Ok(cfg_data.replace("\n",""))
}
pub async fn load_cloud_env(cloud: &mut Cloud, source: &str) -> Result<()> {
  env_cloud(source, &mut cloud.env).await?;
	Ok(())
}
pub async fn load_cloud_config(cloud: &mut Cloud, source: &str) -> Result<(KloudHome,Provider,String), Error> {
	// dbg!(&cloud);
  let cfg_data = load_config_data(&cloud,source).await?;
  let cfg: KloudHome = serde_yaml::from_str(&cfg_data)?;
	let cfg_provider = format!("{}",cfg.provider[0]);
	let provider = cloud.providers.get(&cfg_provider).with_context(||
		 format!("Provider '{}'' not defined", &cfg_provider))?;  
  Ok((cfg, provider.to_owned(), cfg_data))
}
pub async fn load_cloud_check_config(cloud: &mut Cloud, source: &str) -> Result<(KloudCheckHome,Provider,String), Error> {
	// dbg!(&cloud.env);
  let cfg_data = load_config_data(&cloud,source).await?;
  let cfg: KloudCheckHome = serde_yaml::from_str(&cfg_data)?;
	let cfg_provider = format!("{}",cfg.provider[0]);
	let provider = cloud.providers.get(&cfg_provider).with_context(|| format!("Provider '{}'' not defined", &cfg_provider))?;  
  Ok((cfg, provider.to_owned(), cfg_data))
}
pub async fn load_cloud_name_config(cloud: &mut Cloud, source: &str) -> Result<(MainResourcesConfig,Provider,String), Error> {
	// dbg!(&cloud.env);
  let cfg_data = load_config_data(&cloud,source).await?;
  let cfg: MainResourcesConfig = serde_yaml::from_str(&cfg_data)?;
	let provider = cloud.providers.get(&cfg.provider).with_context(|| format!("Provider '{}'' not defined", &cfg.provider))?;  
  Ok((cfg, provider.to_owned(), cfg_data))
}

pub async fn get_cloud_home_list(cloud: &Cloud) -> Result<Vec<String>> {
	let kloud_files: Vec<String> = fs::read_dir(&cloud.env.home)?
	.filter_map(|res| 
	  match res.map(|e| e.file_name()) {
			Ok(entry) => {
				// for e.path()
				// let file_path = entry.as_path().display().to_string();
				let file_path = format!("{}",entry.to_owned().into_string().unwrap_or_else(|_|String::from("")));
				let cfg_path = format!("{}/{}/{}/{}",&cloud.env.home,&file_path,&cloud.env.config_root,&cloud.env.config_path);
				let first_char = file_path.chars().next().unwrap_or_default().to_string(); 
				if first_char.as_str() != "_" && first_char.as_str() != "." &&  Path::new(&cfg_path).exists() { 
					Some(file_path)
				} else {
					None
				}
			},
			Err(e) => {
				eprintln!("Error filter_map {}",e);
				None
			}
		}
	)
	.collect();
	Ok(kloud_files.to_owned())
}

pub fn run_ssh_on_srvr(hostname: &str,tsksrvc_name: &str, tsksrvc_cmd: &str, ssh_access: &SSHAccess) -> Result<String> {
	let debug = envmnt::get_isize("DEBUG",0);
	if debug > 0 {
		println!("Checking connection to {}@{} on {} for {} ",ssh_access.user,ssh_access.host,ssh_access.port,&tsksrvc_name);
		println!("ssh {} /var/lib/klouds/bin/{}_info.sh yaml",&hostname,&tsksrvc_name);
	}
	let output = Command::new("ssh")
		.arg("-q")
		.arg("-o")
		.arg("StrictHostKeyChecking=accept-new")
		.arg("-p")
		.arg(format!("{}",ssh_access.port))
		.arg(format!("{}@{}",ssh_access.user,ssh_access.host))
		.arg("sudo")
		.arg(format!("/var/lib/klouds/bin/{}_info.sh",&tsksrvc_name))
		.arg("yaml")
		.arg(format!("{}",&tsksrvc_cmd))
//		.arg(format!("ssh {} /var/lib/klouds/bin/{}_info.sh yaml",&hostname,&tsksrvc_name))
		.output()?;
	//dbg!(&output);
	if !&output.status.success() {
		return Err(anyhow!("Connection to '{}' for tsksrvc '{}' failed: {}",&hostname,&tsksrvc_name,&output.status));
	} 
	let res = str::from_utf8(&output.stdout).unwrap_or_else(|_| "");
  let info: String = serde_yaml::from_str(&res)
		.unwrap_or_else(|e| {
			eprintln!("serde_yaml: {}",e);
			String::from("")
		});
  Ok(info.to_owned())
}

pub async fn get_provider_info(provider: &str, hostname: &str, cmd: &str , cfg_path: &str) ->  String {
  match provider {
    "upcloud" => {
			get_upcloud_info(hostname,cmd,cfg_path).await
		},
		_ => String::from("")
	}
}
pub fn write_cache_file(output_path: &str,output_data: &str) -> Result<()> {
	if Path::new(&output_path).exists() {
		fs::remove_file(&output_path)?;
	}
	let mut file = OpenOptions::new().write(true).create(true).open(&output_path)?;
	file.write_all(output_data.as_bytes())?;
	Ok(())
}
pub fn read_cache_file(source_path: &str) -> Result<String> {
	let cache_data = fs::read_to_string(&source_path).with_context(|| format!("Failed to read cache from {}",&source_path))?;
	Ok(cache_data)
}
pub async fn create_cloud_config(
	reqname: &str,req_tsksrvcs: &str,reqenv: &ReqEnv,
	 entries: Vec<String>,mut cloud: Cloud
) -> Vec<KloudHome> {
	let config = reqenv.config();
	let debug = envmnt::get_isize("DEBUG",0); 
	let check_path = format!("{}/clouds.json",&config.check_path);
	let	mut check_entries: Vec<KloudCheckHome> = Vec::new();
	let mut no_check_entries = true;
	if reqname != "check_job" {
		if Path::new(&check_path).exists() {
			// Load & Parse reuse liveness and monitor
			let check_data = fs::read_to_string(&check_path).unwrap_or_else(|e|{
				eprintln!("Failed to read 'check_path' from {}: {}", &check_path,e);
				String::from("")
			});
			if !check_data.is_empty() {
				check_entries = serde_json::from_str(&check_data).unwrap_or_else(|e| {
					eprintln!("Error loading check_entries ({}): {}",&check_path,e);
					Vec::new()
				});
				no_check_entries=false;
				if debug> 0 {
					println!("Using check_entries from {}",&check_path);
				}
				// dbg!("{:#?}",&check_entries);
			}
		}
	}
	let mut entries_cfgs: Vec<KloudHome> = Vec::new();
	for (idx, entry) in entries.iter().enumerate() {
		let (mut cfg,_provider,cfg_data) = load_cloud_config(&mut cloud, &entry).await
			.unwrap_or_else(|e| {
				eprintln!("Load_cloud_config {}:",e);
				(KloudHome::default(),Provider::default(),String::from(""))
		});
		if cfg.name.is_empty() || cfg_data.is_empty() {
			let result = format!("Errors loading {}",&entry);
			if debug > 0 {
				println!("{}",&result);
			}
			continue;
		}
	  let entry_cache_path = format!("{}/{}",&config.cache_path,&entry);
		if ! Path::new(&entry_cache_path).exists() {
		 	let dir_path_buf = Path::new(&entry_cache_path).to_path_buf();
		 	let dirs: Vec<&std::path::PathBuf> = vec![&dir_path_buf];
		 	mkdir(&dirs).unwrap_or_else(|e| println!("\nFailed to create dir path {}: {}", &entry_cache_path,e));
		}
		if req_tsksrvcs.contains("monitor") {
		  if no_check_entries {
			  let monitor_cache_path = format!("{}/monitor.json",entry_cache_path);
				let monitor_data = get_cloud_monitor_info(&mut cloud, &entry).await
					.unwrap_or_else(|e| {
						eprintln!("Error {} monitor_info {} -> {}",&entry,&cloud.env.monitor_run,e);
						String::from("")
				});
				write_cache_file(&monitor_cache_path,&monitor_data).unwrap_or_else(|e| println!("Error writing cache {} to {}: {}",&entry,&monitor_cache_path,e));
				cfg.monitor_info = Some(monitor_cache_path);
			} else if check_entries.len() > 0 && check_entries.len() < idx {
				cfg.monitor_info = check_entries[idx].monitor_info.to_owned();
			}
		}
		if reqname.contains("provision") || req_tsksrvcs.contains("resources") || req_tsksrvcs.contains("resources") || req_tsksrvcs.contains("liveness") || req_tsksrvcs.contains("") {
		  let mut groups: Vec<CloudGroup> = Vec::new();
			// cfg.groups = cfg.groups.map(|grp| grp.with_resources(grp.path.to_owned())).collect();
			for (grp_idx,grp) in cfg.groups.iter().enumerate() { 
		    let mut items: Vec<CloudItem> = Vec::new();
			  for (itm_idx,itm) in grp.items.iter().enumerate() { 
					let resources: Option<String>;
					let liveness: Vec<SrvcsHostInfo>;
					let provision: Option<Vec<HostInfo>>;
					if req_tsksrvcs.contains("liveness") {
						if no_check_entries {
							liveness = on_cloud_name_req_check("liveness",&cloud,&reqenv,"","",&itm.path).await;
						} else if check_entries.len() > 0 && check_entries.len() < idx {
							liveness = check_entries[idx].groups[grp_idx].items[itm_idx].liveness.to_owned();
						} else {
							liveness = Vec::new();
						}
					} else {
						liveness = itm.liveness.to_owned();
					}
					if reqname.contains("provision") || req_tsksrvcs.contains("provision") {
						provision = Some(on_cloud_name_req_info("provision",&cloud,&reqenv,"","",&entry,&itm.path).await);
					} else {
						provision = itm.provision.to_owned();
					}
					if req_tsksrvcs.contains("resources") {
			      let resources_cache_path = format!("{}/{}_{}_resources.json",entry_cache_path,&grp.name,&itm.name);
						let resources_data = load_config_json_data(&cloud,&itm.path).await
							.unwrap_or_else(|e| {
								eprintln!("Error loading resources -> {}",e);
								String::from("")
						});
						write_cache_file(&resources_cache_path,&resources_data).unwrap_or_else(|e| println!("Error writing cache {} to {}: {}",&entry,&resources_cache_path,e));
						resources = Some(resources_cache_path);
					} else {
						resources = itm.resources.to_owned();
					}
					items.push(CloudItem {
						name: itm.name.to_owned(),
						info: itm.info.to_owned(),
						path: itm.path.to_owned(),
						resources,
						liveness,
						provision,
					  graph: itm.graph.to_owned(),
						critical: itm.critical.to_owned(),
					});
				}
				groups.push(CloudGroup {
					name: grp.name.to_owned(),
					info: grp.info.to_owned(),
					path: grp.path.to_owned(),
					// TODO check this for group
					resources: grp.resources.to_owned(),
					liveness: grp.liveness.to_owned(),
					items, 
					graph: grp.graph.to_owned(),
					prices: grp.prices.to_owned(),
				});
			}
			cfg.groups=groups;
		} 
		let config_cache_path = format!("{}/cloud_config.json",&entry_cache_path);
		let config_json = serde_json::to_string(&cfg).unwrap_or_else(|e|{ 
			println!("Error creating config_json:{}",e);
			String::from("")
		});
		write_cache_file(&config_cache_path,&config_json).unwrap_or_else(|e| println!("Error writing cache {} to {}: {}",&entry,&config_cache_path,e));
		entries_cfgs.push(cfg.to_owned());
	}
	entries_cfgs
}
pub fn cloud_info_json(clouds: Vec<KloudHome>) -> String {
	let res = "#RES#";
	let mon = "#MON#";
	let res_pat = format!("\"{}\"",&res);
	let mon_pat = format!("\"{}\"",&mon);
	let mut out_clds = String::from("");
	for cld in clouds.iter() {	
		let mut out_grps = String::from("");
		let mut cld_data = cld.to_owned();
	  for (grpidx, grp) in cld.groups.iter().enumerate() { 
			let mut group_data = grp.to_owned();
		  let mut out_items = String::from("");
		  for itm in grp.items.iter() { 
				let resources_data: String;
		    let mut itm_data = itm.to_owned();
				if let Some(res_data) = &itm.resources {
					resources_data = read_cache_file(&res_data).unwrap_or_else(|e|{
						println!("Error reading cache {}: {}",&res_data,e);
			      String::from("\"\"")
					});
				} else {
					resources_data = String::from("\"\"").replace("\\\"","\"");
				}
				itm_data.resources = Some(res.to_owned());
				let mut provision_data = String::from("");
				if let Some(itm_prov) = &itm.provision {
					for (idx,prov) in itm_prov.iter().enumerate() {
						let prov_data = read_cache_file(&prov.info).unwrap_or_else(|e|{
							println!("Error reading cache {}: {}",&prov.info,e);
							String::from("\"\"")
						});
						if idx > 0 {
							provision_data = format!("{},",&provision_data);
						}
						let json_data = format!("{}:{}{}{}{}{}",r#"{"hostname""#,r#"""#,&prov.hostname,r#"","data":["#,&prov_data,r#"]}"#);
					  provision_data = format!("{}{}",&provision_data,&json_data);
					}
				} else {
					provision_data = String::from("\"\"").replace("\\\"","\"");
				}
				itm_data.provision = Some(Vec::new());
				itm_data.info = String::from("");
				out_items = format!("{}{}",&out_items, serde_json::to_string(&itm_data).unwrap_or_else(|_| String::from(""))
				 	.replace("\\\"","\"")
				 	.replace("\n","")
			 	  .replace("\"provision\":[]",&format!("\"provision\":[{}]",&provision_data))
					.replace(&res_pat,&resources_data)
			  );
			}
			if grpidx > 0 {
				out_grps = format!("{},",&out_grps);
			}
			// println!("{}",&out_items);
			group_data.items = Vec::new();
			group_data.info = String::from("");
			out_grps = format!("{}{}",&out_grps, serde_json::to_string(&group_data).unwrap_or_else(|_| String::from(""))
			 	.replace("\"items\":[]",&format!("\"items\":[{}]",&out_items))
		  );
		}
		let mut monitor_data: String;
		if let Some(mon_data) = &cld_data.monitor_info {
			monitor_data = read_cache_file(&mon_data).unwrap_or_else(|e|{
				println!("Error reading cache {}: {}",&mon_data,e);
				String::from("\"\"")
			});
		} else {
			monitor_data = String::from("\"\"");
		}
		if monitor_data.is_empty() {
			monitor_data = String::from("\"\"");
		}
		cld_data.monitor_info = Some(format!("{}",&mon.to_owned()));
		cld_data.groups = Vec::new();
		if !out_clds.is_empty() {
		  out_clds = format!("{},",&out_clds);
		}
		out_clds = format!("{}{}",&out_clds, serde_json::to_string(&cld_data).unwrap_or_else(|_| String::from(""))
			.replace(&mon_pat,&format!("{}",&monitor_data))
			.replace("\"groups\":[]",&format!("\"groups\":[{}]",&out_grps))
			.replace("\\\"","\"")
		);
	}
  out_clds = format!("[{}]",&out_clds).replace("}{","},{");
  out_clds
}